/*
 * Copyright 1995 by the Massachusetts Institute of Technology.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */
/* %%%EOOH%%% */

#include <stdio.h>
#include <stdlib.h>
#include <hesiod.h>
#include <errno.h>
#include "Connect.h"
#include "slw.h"

extern char *pname;
extern void printError();

#define E(x) if (x) printError(1)
#define ED(x) if (x) printError(0)

Addr openConnection(from, who)
     Addr from;
     char *who;
{
  char to[50];
  Addr dest;

  sprintf(to, "%s:%s/%s", PORTDOMAIN, who, PORTNAME);

  if (Connect_NameToAddress(to, &dest))
    {
    errorCheck:
      switch(Error)
	{
	case UDP_UNKNOWNSERVICE:
	  /* fprintf(stderr,
	     "Warning: Service %s not in /etc/services; using port %d\n",
	     PORTNAME, PORTNUMBER); */
	  Error_Pop();
	  sprintf(to, "%s:%s/%d", PORTDOMAIN, who, PORTNUMBER);
	  if (Connect_NameToAddress(to, &dest))
	    goto errorCheck; /* Can't get UNKNOWNSERVICE with explicit port */
	  break;
	case CONNECT_UNKNOWNNAME:
	  ED(CHECK);
	  break;
	default:
	  E(CHECK);
	}
    }

  /*
   * Need discussion of the semantics of OpenConnection
   * here. :-)
   */
  if (dest != NOADDRESS)
    if (Connect_OpenConnection(from, dest))
      {
	if (Error_Severity == S_FATAL)
	  {
	    Connect_FreeAddress(dest);
	    dest = NOADDRESS;
	    ED(CHECK);
	  }
      }

  if (dest == NOADDRESS)
    dest = from;

  return dest;
}

/*
 * Sort a short list of strings.
 * If you want a big list of strings, use qsort.
 * Hey, does qsort(list, len(list), sizeof(char *), strcasecmp)
 * do it? I didn't realize it was so twivial.
 */
sortStrings(list)
     char **list;
{
  int i, j, min;
  char *swap;
  
  if (list[0] == NULL)
    return;

  for (i = 0; list[i + 1] != NULL; i++)
    {
      min = i;
      for (j = i + 1; list[j] != NULL; j++)
	if (strcasecmp(list[min], list[j]) > 0)
	  min = j;

      if (min != i)
	{
	  swap = list[i];
	  list[i] = list[min];
	  list[min] = swap;
	}
    }
}

char **env_resolve()
{
  char *serverenv, *ptr;
  static char *list[MAXSERVERS+1];
  static char *copy;
  int i = 0;

  serverenv = getenv("SLWSERVER");

  if (serverenv == NULL)
    return NULL;

  copy = malloc(strlen(serverenv) + 1);
  if (copy == NULL)
    return NULL;

  ptr = copy;
  list[i++] = ptr;

  while (*serverenv)
    {
      if (*serverenv == ':')
	{
	  *ptr++ = '\0';
	  serverenv++;
	  list[i++] = ptr;
	}
      else
	*ptr++ = *serverenv++;
    }

  *ptr = '\0';

  list[i] = NULL;

  return list;
}

int askQuestionOfServer(req, repl)
     Packet *req, *repl;
{
  static int gotservers = 0, curserver = 0, validaddress = 0;
  static char **servernames;
  static Addr servers[MAXSERVERS];
  int i, startserver, patient;

  if (!gotservers)
    {
      /*
       * Initialize server list.
       */
      gotservers = 1;

      servernames = env_resolve();

      if (servernames == NULL)
	servernames = hes_resolve(SLOCNAME, SLOC);

      if (servernames == NULL)
	{
	  switch(hes_error())
	    {
	    case HES_ER_NOTFOUND:
	      fprintf(stderr, "%s: Hesiod doesn't know about %s.\n",
		      pname, SLOCNAME);
	      break;
	    case HES_ER_CONFIG:
	      fprintf(stderr,
     "%s: Hesiod indicates possible local hesiod configuration problems.\n",
		      pname);
	      break;
	    case HES_ER_NET:
	      fprintf(stderr, "%s: Hesiod reports network problems.\n", pname);
	      break;
	    }

	  fprintf(stderr, "%s: Could not find any servers.\n", pname);
	  exit(1);
	}

      for (i = 0; i < MAXSERVERS; i++)
	servers[i] = NOADDRESS;

      /*
       * We must sort the list of servers because under some versions of
       * bind, the list we get may be randomized. We don't want this because
       * we care, for this service, much more about having most of the logs
       * in a central location than about load balancing.
       */
      sortStrings(servernames); /* XXX Errrr... Shouldn't be doing this. */
    }

  startserver = curserver;

  /*
   * See if we can talk to anyone.
   */
  while (1)
    {
      if (servers[curserver] == NOADDRESS)
	  servers[curserver] = openConnection(req->Source,
					      servernames[curserver]);

      if (servers[curserver] != req->Source)
	{
	  validaddress = 1;
	  req->Destination = servers[curserver];

	  patient = NUMRETRIES;
	  while (patient)
	    {
	      if (Connect_SendPacket(req))
		{
		  if (Error == UDP_SENDTO &&
		      errno == ECONNREFUSED)
		    {
		      /* Got an ICMP port unreachable. No need to
			 timeout. */
		      Error_Pop();
		      break;
		    }
		  else
		    E(CHECK);
		}

	      if (Connect_Wait(repl, RETRY))
		{
		  switch(Error)
		    {
		    case CONNECT_TIMEOUT:
		    case UDP_DEADSOCKET:
		      Error_Pop();
		      patient--;
		      continue;
		      break;

		    /* This isn't actually correct for UDP_DEADSOCKET
		       - this case happens right now under Linux
		       because it is returning info for the ICMP port
		       unreachable packet. I need to update the
		       network library to do the right stuff for this,
		       but for now there's this workaround. This
		       workaround is less incorrect than other more
		       appealing workarounds. */

		    default:
		      ED(CHECK);
		      exit(3);
		    }
		}
	      else
		return 0;
	    }
	}

      curserver++;
      if (servernames[curserver] == NULL)
	{
	  curserver = 0;
	  if (startserver == curserver)
	    return 1;
	}
    }
}
