/*
 * Copyright 1995 by the Massachusetts Institute of Technology.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */
/* %%%EOOH%%% */

/*
 * A program for a slight level of paranoia.
 * Allow programs to be run as other-than-root.
 * Create a .pid file for easy signals.
 *
 * The .pid file is owned by root, not the owner of the process.
 * This way, if the process is insecure, it can't rewrite its own
 * pid file and convince root users using the .pid file to send
 * signals willy nilly.
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

static char *name;

usage()
{
  fprintf(stderr, "usage: %s <owner> <pid file> <path/command...>\n", name);
  exit(1);
}

int main(argc, argv)
     int argc;
     char **argv;
{
  char *owner, *filename, **args;
  FILE *pidfile;
  struct passwd *info;
  uid_t uid;
  gid_t gid;
  pid_t pid;

  name = strrchr(argv[0], '/');
  name = name ? name+1 : argv[0];

  if (argc < 4)
    usage();

  owner = argv[1];
  filename = argv[2];
  args = argv + 3;

  info = getpwnam(owner);
  if (info == NULL)
    {
      fprintf(stderr, "%s: unknown username: %s\n", name, owner);
      exit(1);
    }
  uid = info->pw_uid;
  gid = info->pw_gid;

  (void)umask(022);
  pidfile = fopen(filename, "w");
  if (pidfile == NULL)
    {
      fprintf(stderr, "%s: error (%d) opening %s for writing\n", name,
	      errno, filename);
      exit(1);
    }

  pid = fork();
  if (pid == -1)
    {
      fprintf(stderr, "%s: fork() returned error %d\n", name, errno);
      exit(1);
    }

  /*
   * If parent, write pid file and exit.
   */
  if (pid != 0)
    {
      fprintf(pidfile, "%d\n", pid);
      fclose(pidfile);
      exit(0);
    }

  /*
   * If child, change owner and exec.
   */
  fclose(pidfile);

  if (-1 == setgid(gid))
    {
      fprintf(stderr, "%s: couldn't setgid(%d), aborting\n", name, gid);
      exit(1);
    }

  if (-1 == setuid(uid))
    {
      fprintf(stderr, "%s: couldn't setuid(%d), aborting\n", name, uid);
      exit(1);
    }

  if (-1 == execv(args[0], args))
    {
      fprintf(stderr, "%s: couldn't execv %s\n", name, args[0]);
      exit(1);
    }
}
